from functools import cached_property
from typing import Generator
from typing import Iterable

from explicit.domain.model import asdict
from explicit.messagebus.events import Event
from pydantic.tools import parse_obj_as

from librbac.config import rbac_config
from librbac.types import Permission

from . import logger
from .domain.events import PermissionMarkedObsolete
from .domain.events import PermissionsReplaced
from .types import InvalidStateException
from .types import Migration
from .types import MigrationLoaderInterface
from .types import MigrationRecorderInterface
from .types import PermissionEntry
from .types import PermissionReplacementGraph


class ProjectedSystemState:

    """Прогнозируемое состояние системы, после применения миграций.

    Содержит только разрешения, с которыми выполнялись операции в миграциях.
    """

    _replacements_graph: PermissionReplacementGraph
    _obsolete: set[PermissionEntry]

    def __init__(self):
        self._replacements_graph = PermissionReplacementGraph()
        self._obsolete = set()

    def handle_replacement(self, replacements: list[PermissionEntry], replaces: list[PermissionEntry]):
        """Обработка замены разрешений."""
        if obsolete_replacements := self._obsolete.intersection(set(replacements)):
            raise InvalidStateException(f'Замены уже были помечены как устаревшие: {obsolete_replacements}.')
        if obsolete_replaced := self._obsolete.intersection(set(replaces)):
            raise InvalidStateException(f'Заменяемые разрешения уже были помечены как устаревшие: {obsolete_replaced}')

        self._replacements_graph.add_replacements(set(replacements), set(replaces))

    def handle_obsolescence(self, *obsolete: PermissionEntry):
        """Обработка устаревания разрешений."""
        obsolete = set(obsolete)
        if already_obsolete := obsolete.intersection(self._obsolete):
            raise InvalidStateException(f'Разрешения уже были помечены как устаревшие: {already_obsolete}')

        if already_replaced := obsolete.intersection(self._replacements_graph.internal_nodes):
            raise InvalidStateException(f'Разрешения уже были помечены как заменённые: {already_replaced}')

        self._obsolete.update(obsolete)

    def _entry_to_rbac_permission(self, permission_entry: PermissionEntry) -> Permission:
        return Permission(
            namespace=permission_entry.namespace,
            resource=permission_entry.resource,
            action=permission_entry.action,
            scope=permission_entry.scope
        )

    @property
    def available_permission_entries(self) -> set[Permission]:
        return {
            self._entry_to_rbac_permission(entry)
            for entry in self._replacements_graph.leaf_nodes - self._obsolete
        }

    @property
    def replaced_permission_entries(self) -> set[Permission]:
        return {
            self._entry_to_rbac_permission(entry)
            for entry in self._replacements_graph.internal_nodes
        }

    @property
    def obsolete_permission_entries(self) -> set[Permission]:
        return {
            self._entry_to_rbac_permission(entry)
            for entry in self._obsolete
        }


class EventGenerator:

    """Генератор набора событий из миграции."""

    _migration: Migration

    def __init__(self, migration: Migration):
        self._migration = migration

    def __iter__(self) -> 'Generator[Event]':
        for replaced_entry in self._migration.replaced:
            yield parse_obj_as(PermissionsReplaced, asdict(replaced_entry))
        for obsolete_entry in self._migration.obsolete:
            yield PermissionMarkedObsolete(permission=asdict(obsolete_entry))


class MigrationOrchestrator:
    """Выполняет общее управление процессом сборки и применения миграций."""

    _loader: MigrationLoaderInterface
    _recorder: MigrationRecorderInterface

    def __init__(self, loader: type[MigrationLoaderInterface], recorder: type[MigrationRecorderInterface]):
        self._loader = loader()
        self._recorder = recorder()

    @cached_property
    def _migrations(self) -> tuple[tuple[str, str, Migration], ...]:
        """Возвращает миграции."""
        return tuple(
            (app_label, migration_name, migration)
            for app_label, migration_name, migration in self._loader.get_migrations()
            if (app_label, migration_name) not in self._recorder.applied_migrations
        )

    def _build_projected_state(self, migrations: Iterable[Migration]) -> ProjectedSystemState:
        """Воссоздаёт ожидаемое состояние системы из миграций."""
        projected_system_state = ProjectedSystemState()
        for migration in migrations:
            for op in migration.replaced:
                projected_system_state.handle_replacement(replacements=op.replacements, replaces=op.replaces)
            projected_system_state.handle_obsolescence(*migration.obsolete)
        return projected_system_state

    def _verify_projected_state(self, projected_system_state: ProjectedSystemState):
        """Проверяет на противоречие состояния и кодовой базы."""
        from librbac.manager import rbac

        rbac_permissions = set(rbac.permissions)
        # убеждаемся, что устаревшие разрешения из миграций отсутствуют в системе
        if obsolete_in_rbac := projected_system_state.obsolete_permission_entries.intersection(rbac_permissions):
            raise Exception(f'Помеченный устаревшими разрешения всё еще существуют в системе: {obsolete_in_rbac}')
        # убеждаемся, что заменённые в миграциях разрешения отсутствуют в системе
        if replaced_in_rbac := projected_system_state.replaced_permission_entries.intersection(rbac_permissions):
            raise Exception(f'Помеченный заменёнными разрешения всё еще существуют в системе: {replaced_in_rbac}')
        # убеждаемся, что замены из миграций присутствуют в системе
        if replacements_not_in_rbac := projected_system_state.available_permission_entries.difference(rbac_permissions):
            Exception(f'Замены из миграций отсутствуют в системе: {replacements_not_in_rbac}')

    def apply_migrations(self, dry_run=False):
        """Применяет миграции."""

        logger.info('Формирование из миграций ожидаемого состояния системы...')
        projected_state: ProjectedSystemState = self._build_projected_state(
            tuple((migration for _, _, migration in self._migrations))
        )

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        logger.info(f'{len(projected_state.obsolete_permission_entries)} разрешений помечены как устаревшие.')
        for entry in projected_state.obsolete_permission_entries:
            logger.info(f'\t{entry}')

        logger.info(f'{len(projected_state.replaced_permission_entries)} разрешений заменены.')
        for entry in projected_state.replaced_permission_entries:
            logger.info(f'\t{entry}')
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self._verify_projected_state(projected_state)
        logger.info('Выгрузка событий...')
        for app_label, migration_name, migration in self._migrations:
            for event in EventGenerator(migration):
                if not dry_run:
                    rbac_config.bus.handle(event)
                logger.info(f'\t{event}')
            if not dry_run:
                self._recorder.record_applied(app=app_label, name=migration_name)



from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from explicit.adapters.messaging import AbstractAdapter

from ..domain.events import PermissionMarkedObsolete
from ..domain.events import PermissionsReplaced


def publish_to_adapter(event: PermissionMarkedObsolete | PermissionsReplaced, messaging_adapter: 'AbstractAdapter'):
    """Обработчик событий миграций разрешений, публикующий событий в межсервисную шину."""
    from librbac.utils.rbac import get_rbac_topic_permission
    messaging_adapter.publish(get_rbac_topic_permission(), event.dump())



from importlib import import_module
from typing import TYPE_CHECKING
from typing import Generator
from typing import Set
from typing import Type
from typing import Union

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from rest_framework.viewsets import GenericViewSet


if TYPE_CHECKING:
    from librbac.drf.viewsets import RBACMixin


def get_rbac_topic_permission() -> str:
    from django.conf import settings
    topic_attr = 'RBAC_PERMISSION_TOPIC'
    from librbac import default_settings
    topic_name = getattr(settings, topic_attr, None) or getattr(default_settings, topic_attr)
    if not topic_name:
        raise ImproperlyConfigured(
            f'Необходимо указать {topic_attr} в настройках.'
        )
    return topic_name


def get_all_viewsets(
    urlpatterns,
    viewsets=None
) -> Set[Type[GenericViewSet]]:
    """Находит и возвращает все ViewSet'ы системы."""
    viewsets = viewsets if viewsets is not None else set()
    for pattern in urlpatterns:
        if hasattr(pattern, 'url_patterns'):
            get_all_viewsets(pattern.url_patterns, viewsets=viewsets)
        else:
            if cls := getattr(pattern.callback, 'cls', None):
                if issubclass(cls, GenericViewSet):
                    viewsets.add(cls)

    return viewsets


def _get_rbac_viewsets() -> Generator[Union[GenericViewSet, 'RBACMixin'], None, None]:
    """Возвращает ViewSet'ы системы, доступ к которым нужно проверять."""
    from librbac.drf.viewsets import RBACMixin
    urlpatterns = import_module(settings.ROOT_URLCONF).urlpatterns
    for viewset in get_all_viewsets(urlpatterns):
        if issubclass(viewset, RBACMixin):
            yield viewset



from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING
from typing import Iterable
from typing import NamedTuple
from typing import Optional
from typing import Tuple
from typing import TypedDict
from typing import Union


if TYPE_CHECKING:
    from rest_framework.request import Request
    from rest_framework.viewsets import GenericViewSet


class PermissionGroup(NamedTuple):

    """Структура группы разрешений."""

    namespace: str
    resource: str


class Permission(NamedTuple):

    """Структура разрешения."""

    namespace: str
    resource: str
    action: str
    scope: Optional[str] = None


class TokenPermissionDict(TypedDict):

    """Структура разрешения передаваемого в составе токена."""

    resource_set_id: str
    """Объект доступа, например `auth:role`"""
    scopes: Iterable[str]
    """Операции на которые пользователю выдан доступ, например `read`"""
    exp: int
    """Временная метка после которой набор прав должен быть обновлён"""


TUser = Union['AbstractBaseUser', 'OIDCUser']
"""Тип "Пользователь системы"."""


class TRule(ABC):
    """Тип "Правило для проверки доступа"."""
    @abstractmethod
    def __call__(
        self,
        viewset: 'GenericViewSet',
        request: 'Request',
        user: TUser,
    ) -> bool:
        """Проверка доступа."""


TPermMapDict = dict[str, Tuple[Permission, ...]]
"""Структура сопоставления разрешений с действиями во ViewSet'е."""




from django.db import models


class MigrationRecord(models.Model):

    app = models.CharField(max_length=255)
    name = models.CharField(max_length=255)
    applied = models.DateTimeField(auto_now_add=True)




from django.core.management.base import BaseCommand
from django.core.management.base import CommandError

from librbac.contrib.migrations.loader.disk import MigrationLoader
from librbac.contrib.migrations.orchestrator import MigrationOrchestrator
from librbac.contrib.migrations.recorder.local_db import MigrationRecorder
from librbac.manager import rbac


class Command(BaseCommand):

    help = 'Выполнить миграцию разрешений'

    def add_arguments(self, parser):
        parser.add_argument(
            '--dry-run',
            action='store_true',
            help=(
                'Пробный запуск миграций. Выполняется только сборка и проверка целостности. '
                'Обработка событий не выполняется.'
            )
        )

    def handle(self, *args, **options):
        try:
            from librbac.config import rbac_config
        except ImportError as ie:
            raise CommandError(
                'Не удалось импортировать настройки модуля. '
                'Вероятно, система контроля доступа не сконфигурирована.'
            ) from ie

        rbac.init()
        orchestrator = MigrationOrchestrator(loader=MigrationLoader, recorder=MigrationRecorder)
        orchestrator.apply_migrations(dry_run=options['dry_run'])



from importlib import import_module
from itertools import chain
from itertools import product
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING
from typing import Generator
from typing import Tuple

from django.apps import apps
from pydantic.tools import parse_file_as

from ..types import Migration
from ..types import MigrationLoaderInterface


if TYPE_CHECKING:
    from django.apps.config import AppConfig


class MigrationLoader(MigrationLoaderInterface):

    """Загрузчик миграций с диска"""

    _disk_migrations: dict[tuple[str, str], Migration]

    def __init__(self):
        self._disk_migrations = {}

        self._load_disk()

    @classmethod
    def _get_app_permissions_modules(cls) -> Generator[Tuple[ModuleType, 'AppConfig'], None, None]:
        """Возвращает модули permissions из приложений системы."""
        for app_config in apps.get_app_configs():
            try:
                yield import_module('.permissions', app_config.name), app_config
            except ImportError as error:
                if 'No module named' not in error.args[0]:
                    raise
                continue

    def _load_disk(self):
        """Load the migrations from all INSTALLED_APPS from disk."""

        for permission_module, app_config in self._get_app_permissions_modules():
            migrations_dir_path = Path(permission_module.__file__).parent / 'migrations'
            if not migrations_dir_path.is_dir():
                continue

            for migration_file_path in sorted(migrations_dir_path.iterdir(), key=lambda fn: fn.name):
                # Check if the file name starts with a number
                if not (migration_file_path.name[:4].isdigit() and migration_file_path.is_file()):
                    raise ValueError(f'Имя файла миграции не соответствует требованиям: {migration_file_path.name}')

                self._disk_migrations[
                    app_config.label, migration_file_path.name
                ] = parse_file_as(Migration, migration_file_path)

    def get_migrations(self) -> Generator[tuple[str, str, Migration], None, None]:
        for (app_label, migration_name), migration_obj in self._disk_migrations.items():
            yield app_label, migration_name, migration_obj



from typing import TYPE_CHECKING

from django.utils.functional import cached_property

from librbac.backends.base import BackendBase
from librbac.events import PermissionPushed


if TYPE_CHECKING:
    from librbac.events import PermissionPushed


class RemoteBackend(BackendBase):

    """Бэкенд RBAC с публикацией разрешений в межсервисную шину."""

    @cached_property
    def _rbac_topic_permission(self) -> str:
        from librbac.utils.rbac import get_rbac_topic_permission
        return get_rbac_topic_permission()

    def push_permission(self, event: 'PermissionPushed'):
        from librbac.config import rbac_config
        rbac_config.adapter.publish(
            self._rbac_topic_permission,
            event.dump()
        )




from django.apps.config import AppConfig as AppConfigBase


class AppConfig(AppConfigBase):

    name = __package__

    def _register_events(self):
        """Регистрация событий топика."""
        from librbac.config import rbac_config
        from librbac.events import PermissionPushed
        from librbac.utils.rbac import get_rbac_topic_permission
        rbac_config.event_registry.register(
            get_rbac_topic_permission(),
            PermissionPushed,
        )

    def ready(self):
        self._register_events()



from typing import Generator
from typing import Iterable
import json

from jwcrypto import jwt

from librbac.types import Permission
from librbac.types import TokenPermissionDict


def get_jwt_permissions(
    perms: Iterable[TokenPermissionDict]
) -> Generator[Permission, None, None]:
    """Генератор разрешений пользователя.

    Преобразует разрешения словаря в объект разрешений.
    Возвращает генератор объектов разрешений.
    """
    for perm in perms:
        namespace, resource = perm['resource_set_id'].split(':')
        for action_scope in perm['scopes']:
            action, scope = (
                action_scope.split(':')
                if ':' in action_scope
                else (action_scope, None)
            )
            yield Permission(namespace, resource, action, scope)


def set_current_user_permissions(token: str, request):
    """
    Извлекает из токена и присваивает пользователю информацию о разрешениях.
    """
    jwt_obj = jwt.JWT(jwt=token, key=request.client.jwk_key)
    claims = json.loads(jwt_obj.claims)
    perms: Iterable[TokenPermissionDict] = claims.get('permissions', [])
    request.user.permissions = list(get_jwt_permissions(perms))



from collections import defaultdict
from importlib import import_module
from itertools import chain
from types import ModuleType
from typing import TYPE_CHECKING
from typing import Dict
from typing import Generator
from typing import Iterable
from typing import Set
from typing import Tuple

from django.apps import apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.functional import cached_property

from librbac.types import Permission
from librbac.types import PermissionGroup
from librbac.types import TRule
from librbac.utils.rbac import _get_rbac_viewsets


if TYPE_CHECKING:

    from rest_framework.viewsets import GenericViewSet

from contextlib import suppress


def _get_app_permissions_modules() -> Generator[ModuleType, None, None]:
    """Возвращает модули permissions из приложений системы."""
    for app_config in apps.get_app_configs():
        with suppress(ModuleNotFoundError):
            yield import_module('.permissions', app_config.name)


class RBACManager:

    """Менеджер системы контроля доступа RBAC."""

    _msg_perm_not_found = 'Разрешение "{}" не востребовано'

    def __init__(self):
        self.partitions: Dict[str, set[PermissionGroup]] = defaultdict(set)  # Разделы системы
        self.permissions: Dict[Permission, str] = {}
        self.permission_dependencies: Dict[Permission, Set[Permission, ...]] = defaultdict(set)
        self.permission_rules: dict[Permission, list[TRule]] = defaultdict(list)

    def _collect_partitions(self, permission_modules: Tuple[ModuleType, ...]):
        """Сбор информации о разделах системы.

        Разделами системы являются объединения групп разрешений.
        """
        self.partitions.clear()

        processed_groups = set()

        for module in permission_modules:
            partitions = getattr(module, 'partitions', None)
            if partitions is None:
                continue

            for title, perm_groups in partitions.items():
                for perm_group in perm_groups:
                    assert perm_group not in processed_groups, (
                        f'Группа разрешений "{perm_group}" уже '
                        f'закреплена за другим разделом системы.'
                    )
                    self.partitions[title].add(perm_group)
                    processed_groups.add(perm_group)

    def _collect_permissions(
        self,
        permissions_modules: Tuple[ModuleType, ...]
    ):
        """Сбор разрешений системы."""
        self.permissions.clear()

        # Сбор названий разрешений всей системы
        for viewset in _get_rbac_viewsets():
            for permission in chain.from_iterable(viewset.perm_map.values()):
                self.permissions[permission] = ''

        # Заполнение параметров разрешений
        for module in permissions_modules:
            for params in getattr(module, 'permissions', ()):
                if len(params) == 2:
                    permission, title = params
                else:
                    raise ValueError(
                        f'Некорректные параметры разрешения:  {repr(params)}'
                    )

                assert permission in self.permissions, (
                    self._msg_perm_not_found.format(permission)
                )

                self.permissions[permission] = (
                    self.permissions[permission] or title or ''
                )

    def _collect_dependencies(
        self,
        permissions_modules: Tuple[ModuleType, ...]
    ):
        """Сбор зависимостей между разрешениями."""
        self.permission_dependencies.clear()

        for module in permissions_modules:
            module_dependencies = getattr(module, 'dependencies', {})
            if callable(module_dependencies):
                module_dependencies = module_dependencies()

            for permission, dependencies in module_dependencies.items():
                assert permission in self.permissions, (
                    self._msg_perm_not_found.format(permission)
                )

                self.permission_dependencies[permission].update(dependencies)

    def _collect_rules(self, permissions_modules: Tuple[ModuleType, ...]):
        """Сбор обработчиков правил для разрешений системы."""
        self.permission_rules.clear()

        for module in permissions_modules:
            for permission, handlers in getattr(module, 'rules', {}).items():
                if not isinstance(handlers, Iterable):
                    handlers = (handlers,)

                for handler in handlers:
                    assert callable(handler), handler
                    self.permission_rules[permission].append(handler)

    def _push_permissions(self):
        """
        Обновление списка разрешений в хранилище основе разрешений системы.
        """
        self._backend.push_permissions()

    def get_dependent_permissions(
        self,
        permission: 'Permission',
        _result: Set = None
    ) -> Set[Permission]:
        """Возвращает разрешения, от которых зависит указанное разрешение."""
        if _result is None:
            primary_permission = permission
            _result = {permission}
        else:
            primary_permission = None

        for dependency in self.permission_dependencies[permission]:
            if dependency not in _result:
                _result.add(dependency)
                _result.update(
                    self.get_dependent_permissions(dependency, _result)
                )

        if primary_permission:
            _result.remove(primary_permission)

        return _result

    def init(self, push_permissions: bool = False):
        """Инициализация модуля контроля доступа.

        1. Загружает из приложений системы списки правил и разрешений. Их поиск
           осуществляется в модуле ``permissions``.
        2. Для каждого правила и разрешения создает/обновляет запись
            в хранилище.

        :param bool push_permissions: Определяет необходимость синхронизации
            прав доступа системы с хранилищем.
        """
        modules = tuple(_get_app_permissions_modules())

        # Сбор разрешений системы и зависимостей между ними
        self._collect_permissions(modules)
        self._collect_dependencies(modules)

        # Сбор обработчиков правил для разрешений системы
        self._collect_rules(modules)

        # Сбор групп разрешений и разделов системы
        self._collect_partitions(modules)

        if push_permissions:
            self._push_permissions()

    @cached_property
    def _backend(self):
        from librbac import default_settings
        backend_name = getattr(settings, 'RBAC_BACKEND', None) or getattr(default_settings, 'RBAC_BACKEND')
        module_name, class_name = backend_name.rsplit('.', 1)

        try:
            module = import_module(module_name)
        except ImportError as e:
            module = import_module(module_name)
            raise ImproperlyConfigured(
                f'Не удалось импортировать бэкенд RBAC {module_name}: "{e}"'
            ) from e

        try:
            backend_class = getattr(module, class_name)
        except AttributeError as e:
            raise ImproperlyConfigured(
                f'Модуль "{module}" не содержит бэкенд RBAC"{class_name}"'
            ) from e
        else:
            backend = backend_class(self)

        return backend

    def has_access(self, viewset: 'GenericViewSet', request) -> bool:
        """Проверяет наличие у пользователя доступа (с учётом правил)."""
        return self._backend.has_access(viewset, request)

    def has_perm(self, user, permission: Permission) -> bool:
        """Проверяет наличие у пользователя разрешения (без учёта правил)."""
        return self._backend.has_perm(user, permission)


rbac = RBACManager()



from ..models import MigrationRecord
from ..types import MigrationRecorderInterface


class MigrationRecorder(MigrationRecorderInterface):

    @property
    def migration_qs(self):
        return MigrationRecord.objects.all()

    @property
    def applied_migrations(self) -> tuple[tuple[str, str]]:
        return tuple(
            (migration.app, migration.name)
            for migration in self.migration_qs
        )

    def record_applied(self, app: str, name: str):
        """Record that a migration was applied."""
        self.migration_qs.create(app=app, name=name)




from typing import TYPE_CHECKING
from typing import Union

from django.utils.decorators import classonlymethod

from librbac.drf.permissions import HasAccess
from librbac.types import Permission
from librbac.types import TPermMapDict


if TYPE_CHECKING:
    from rest_framework.viewsets import GenericViewSet


class RBACMixin:

    """Примесь для ViewSet'ов требующих контроля доступа."""

    perm_map: TPermMapDict
    """Сопоставление ViewSet'а и требуемых разрешений."""

    permission_classes = (
        HasAccess,
    )

    @classonlymethod
    def as_view(cls: Union['GenericViewSet', 'RBACMixin'], actions=None, **initkwargs):
        assert isinstance(cls.perm_map, dict)
        assert len(cls.perm_map)
        for action, perms in cls.perm_map.items():
            assert all(isinstance(p, Permission) for p in perms)
        return super().as_view(actions=actions, **initkwargs)



from typing import TYPE_CHECKING

from explicit.messagebus.events import Event
from pydantic import Field
from pydantic import conlist

from librbac.events import Permission


if TYPE_CHECKING:
    from dataclasses import dataclass  # noqa
else:
    from pydantic.dataclasses import dataclass  # noqa



@dataclass
class PermissionsReplaced(Event):
    """Разрешения были заменены."""
    replacements: conlist(Permission, min_items=1) = Field(default_factory=list)
    replaces: conlist(Permission, min_items=1) = Field(default_factory=list)


@dataclass
class PermissionMarkedObsolete(Event):
    """Разрешение было помечено как устаревшее."""
    permission: Permission = Field()



from typing import TYPE_CHECKING
from typing import Optional
from typing import Union

from explicit.domain.model import Unset
from explicit.domain.model import unset
from explicit.messagebus.events import Event


if TYPE_CHECKING:
    from dataclasses import dataclass  # noqa
else:
    from pydantic.dataclasses import dataclass  # noqa


@dataclass
class Permission:
    """Данные разрешения."""
    namespace: Union[str, Unset] = unset
    resource: Union[str, Unset] = unset
    action: Union[str, Unset] = unset
    scope: Optional[str] = None
    title: Union[str, Unset] = unset
    module: Union[str, Unset] = unset


@dataclass
class PermissionPushed(Permission, Event):
    ...



from abc import ABC
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from explicit.contrib.adapters.messaging.kafka import Adapter
    from explicit.contrib.messagebus.event_registry import Registry
    from explicit.messagebus import MessageBus


class IConfig(ABC):
    """Конфигурация модуля управления доступом на основе ролей."""

    bus: 'MessageBus'
    """Внутренняя шина сервиса."""

    adapter: 'Adapter'
    """Адаптер к kafka"""

    event_registry: 'Registry'
    """Реестр событий по топикам."""


rbac_config: IConfig



from librbac.backends.base import BackendBase
from librbac.events import PermissionPushed


class LocalBackend(BackendBase):

    """Бэкенд RBAC с публикацией разрешений в локальную шину."""

    def push_permission(self, event: 'PermissionPushed'):
        from librbac.config import rbac_config
        rbac_config.bus.handle(event)



from django.core.management.base import BaseCommand
from django.core.management.base import CommandError

from librbac.manager import rbac


class Command(BaseCommand):

    def add_arguments(self, parser):
        parser.add_argument(
            '--list-permissions', action='store_true', help='Напечатать список загруженных разрешений'
        )
        parser.add_argument(
            '--push-permissions', action='store_true', help='Сохранить загруженные разрешения в хранилище'
        )

    def _write_line(self, namespace, resource, action, scope, title):
        self.stdout.write(
            '{}\t\t{}\t\t{}\t\t{}\t\t{}\n'.format(namespace, resource, action, scope, title)
        )

    def handle(self, *args, **options):
        try:
            from librbac.config import rbac_config
        except ImportError as ie:
            raise CommandError(
                'Не удалось импортировать настройки модуля. '
                'Вероятно, система контроля доступа не сконфигурирована.'
            ) from ie

        if options['list_permissions']:
            rbac.init()
            self._write_line('namespace', 'resource', 'action', 'scope', 'title')
            for (namespace, resource, action, scope), title in rbac.permissions.items():
                self._write_line(namespace, resource, action, scope, title)
        if options['push_permissions']:
            self.stdout.write('Сохранение загруженных разрешений в хранилище...')
            rbac.init(push_permissions=True)
            self.stdout.write('Завершено.')



from importlib import import_module
from inspect import currentframe

from django.core.management import ManagementUtility


def is_in_management_command() -> bool:
    """
    Возвращает True, если код выполняется в рамках миграций Django.
    """
    from django.core.management import ManagementUtility

    def is_in_command(command):
        frame = currentframe()
        while frame:
            if 'self' in frame.f_locals:
                self_object = frame.f_locals['self']
                if isinstance(self_object, command):
                    return True

                if isinstance(self_object, ManagementUtility):
                    # Срабатывает при использовании функции в AppConfig
                    if 'subcommand' in frame.f_locals:
                        subcommand = frame.f_locals['subcommand']
                        return subcommand in ['migrate', 'test']

            frame = frame.f_back

    modules = (
        'django.core.management.commands.migrate',
        'django.core.management.commands.makemigrations',
        'django.core.management.commands.sqlmigrate',
        'django.core.management.commands.showmigrations',
        'django.core.management.commands.test',
    )

    for module_name in modules:
        if is_in_command(import_module(module_name).Command):  # type: ignore
            return True

    return False



from abc import ABCMeta
from abc import abstractmethod
from itertools import product
from typing import TYPE_CHECKING
from typing import Generator

from pydantic import BaseModel
from pydantic import conlist
from pydantic import root_validator


if TYPE_CHECKING:
    from dataclasses import dataclass  # noqa



else:
    from pydantic.dataclasses import dataclass  # noqa


@dataclass(frozen=True)
class PermissionEntry:

    """Запись о разрешении в миграции."""

    namespace: str
    resource: str
    action: str
    title: str
    module: str
    scope: str | None = None


@dataclass(frozen=True)
class ReplacedEntry:

    """Запись о замене разрешений в миграции."""

    replacements: conlist(PermissionEntry, min_items=1)
    replaces: conlist(PermissionEntry, min_items=1)


class Migration(BaseModel):

    """Миграция разрешений."""

    description: str
    replaced: conlist(ReplacedEntry, min_items=0)
    obsolete: conlist(PermissionEntry, min_items=0)

    @root_validator
    def check_has_operations(cls, values):
        assert values['replaced'] or values['obsolete'], 'Миграция должна содержать хотя-бы одну операцию'
        return values


class InvalidStateException(BaseException):
    """Нарушена целостность состояния."""


class PermissionReplacementGraph:

    """Хранилище замен разрешений в виде графа."""

    def __init__(self):
        self._graph = {}

    def _add_permission(self, entry: PermissionEntry):
        """Добавляет в граф новый узел разрешения, если он еще не существует."""
        if entry not in self._graph:
            self._graph[entry] = set()

    def add_replacements(self, replacements: set[PermissionEntry], replaces: set[PermissionEntry]):
        """Добавляет замену разрешений."""

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        if invalid_replacements := self.internal_nodes.intersection(replacements):
            raise InvalidStateException(f'Некорректная замена: {invalid_replacements}')
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        for replacement, replaced in product(replacements, replaces):
            self._add_permission(replaced)
            self._add_permission(replacement)
            self._graph[replaced].add(replacement)

    def __contains__(self, item: PermissionEntry):
        return item in self._graph

    def __getitem__(self, item: PermissionEntry) -> set[PermissionEntry]:
        """Получение заменяемых разрешений по экземпляру замены."""
        return {entry for entry, replaced_by in self._graph.items() if item in replaced_by}

    @property
    def internal_nodes(self) -> set[PermissionEntry]:
        """Набор разрешений, которые должны быть удалены из системы после применения миграций."""
        return {node for node in self._graph if self._graph[node]}

    @property
    def leaf_nodes(self) -> set[PermissionEntry]:
        """Набор разрешений, которые должны быть в системе после применения миграций."""
        return {node for node in self._graph if not self._graph[node]}


class MigrationRecorderInterface(metaclass=ABCMeta):
    """Интерфейс регистратора миграций."""

    @property
    @abstractmethod
    def applied_migrations(self) -> tuple[tuple[str, str]]:
        """Возвращает сопоставление всех применённых миграций по приложениям."""

    @abstractmethod
    def record_applied(self, app: str, name: str):
        """Регистрирует миграцию как применённую."""


class MigrationLoaderInterface(metaclass=ABCMeta):
    """Интерфейс загрузчика миграций разрешений."""

    @abstractmethod
    def get_migrations(self) -> Generator[tuple[str, str, Migration], None, None]:
        """Возвращает все доступные миграции."""





from django.apps.config import AppConfig as AppConfigBase


class AppConfig(AppConfigBase):

    name = __package__
    label = 'librbac_migrations'

    def _register_events(self):
        """Регистрация событий топика."""
        from librbac.config import rbac_config
        from librbac.contrib.migrations.domain.events import (
            PermissionMarkedObsolete)
        from librbac.contrib.migrations.domain.events import (
            PermissionsReplaced)
        from librbac.utils.rbac import get_rbac_topic_permission
        rbac_config.event_registry.register(
            get_rbac_topic_permission(),
            PermissionMarkedObsolete,
            PermissionsReplaced,
        )

    def ready(self):
        self._register_events()



from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Tuple
from typing import Union
from uuid import UUID


if TYPE_CHECKING:
    from librbac.types import Permission


@dataclass
class OIDCUser:
    pk: Union[str, UUID]
    aud: str
    client_id: str
    exp: int
    iat: int
    iss: str
    sub: Union[str, UUID]
    uid: Union[str, UUID]
    permissions: Tuple['Permission']

    # пользователь по-умолчанию считается аутентифицированным
    is_authenticated = True


def get_user_from_token(request, id_token) -> OIDCUser:
    """Возвращает объект пользователя с разрешениями."""
    from librbac.utils.jwt import get_jwt_permissions

    token_data = {**id_token, 'pk': id_token['uid']}
    permissions = tuple(get_jwt_permissions(token_data.pop('permissions')))
    return OIDCUser(**token_data, permissions=permissions)



from typing import TYPE_CHECKING

from rest_framework.permissions import BasePermission

from librbac.manager import rbac


class HasAccess(BasePermission):

    def has_permission(self, request, view):
        """Проверяет наличие доступа."""
        from librbac.drf.viewsets import RBACMixin
        if not isinstance(view, RBACMixin):
            return True
        return rbac.has_access(view, request)



from abc import ABCMeta
from abc import abstractmethod
from typing import TYPE_CHECKING
from typing import Generator
from typing import Tuple

from librbac.events import PermissionPushed
from librbac.manager import RBACManager
from librbac.types import Permission
from librbac.types import PermissionGroup


if TYPE_CHECKING:
    from rest_framework.viewsets import GenericViewSet


class BackendBase(metaclass=ABCMeta):

    """Базовый класс для бэкендов RBAC."""

    def __init__(self, manager):
        self._manager: RBACManager = manager

    def _need_check_access(self, viewset: 'GenericViewSet') -> bool:
        """Возвращает True, если ``viewset`` предполагает проверку доступа."""
        result = False
        if perm_map := getattr(viewset, 'perm_map'):
            result = viewset.action in perm_map
        return result

    def _get_current_user(self, request):
        """Возвращает текущего пользователя."""
        return request.user

    def _get_user_permissions(self, user) -> Tuple['Permission', ...]:
        """Возвращает все доступные пользователю разрешения."""
        return getattr(user, 'permissions', ())

    def _get_viewset_permissions(
        self,
        viewset: 'GenericViewSet'
    ) -> Tuple[Permission]:
        """Возвращает имена разрешений ViewSet'а."""
        return viewset.perm_map.get(viewset.action, ())

    def _check_permission(
        self,
        permission: 'Permission',
        viewset: 'GenericViewSet',
        request,
        user,
    ):
        """Проверяет возможность предоставления доступа.

        Если для указанного разрешения определены правила, то выполняет их
        проверку.
        """
        if permission in self._manager.permission_rules:
            for handler in self._manager.permission_rules[permission]:
                if handler(viewset, request, user):
                    result = True
                    break
            else:
                result = None
        else:
            # Для разрешения не определено правил, значит достаточно
            # только наличия у пользователя разрешения как такового.
            result = True

        return result

    def _get_push_events(self) -> Generator[PermissionPushed, None, None]:
        """Генерирует события для обновления разрешений в хранилище."""
        module_perm_group_mapping = {
            perm_group: module
            for module, perm_groups in self._manager.partitions.items()
            for perm_group in perm_groups
        }
        for permission, title in self._manager.permissions.items():
            yield PermissionPushed(
                namespace=permission.namespace,
                resource=permission.resource,
                action=permission.action,
                scope=permission.scope,
                title=title,
                module=module_perm_group_mapping[PermissionGroup(permission.namespace, permission.resource)],
            )

    @abstractmethod
    def push_permission(self, event: PermissionPushed):
        """Обновляет одно разрешение в хранилище."""

    def push_permissions(self):
        """
        Обновление списка разрешений в хранилище основе разрешений системы.
        """
        for event in self._get_push_events():
            self.push_permission(event)

    def has_access(self, viewset: 'GenericViewSet', request) -> bool:
        """
        Проверяет наличие у текущего пользователя доступа (включая правила).
        """
        if not self._need_check_access(viewset):
            return True

        user = self._get_current_user(request)
        if user is None:
            return False

        return any(
            self._check_permission(permission, viewset, request, user)
            for permission in self._get_user_permissions(user)
            if permission in self._get_viewset_permissions(viewset)
        )

    def has_perm(self, user, permission: 'Permission') -> bool:
        """Проверяет наличие у пользователя разрешения."""
        return permission in self._get_user_permissions(user)
